"""
A collection of utilities for declaring schemas using type annotations.
"""
from .format_registry import FormatRegistry
from .types import setup
from .object import Object
from .formats import Object as ObjectFormat, Dict as DictFormat


setup()


def schema(model=None, **constraints):
    """
    Build a schema. Type annotations on the schema are used to declare the
    schema's item types. Attributes values on the schema are used as default
    values. The new format is used when the model is loaded, dumped, or nested
    in another schema.

    Arguments:
        model (type, optional): The model that the schema will load to and dump
            from. Use this to specify the model if the schema and the model are
            separate classes. Defaults to using the schema as the model.
        ignore_extra (bool, optional): Discard extra values without erroring?
            Defaults to `False`.
        keep_extra (bool, optional): Keep all values even if they are not
            declared by the schema annotations? Overrides ``ignore_extra``.
            Defaults to `False`. *Note: Enabling this on object models
            requires the model's constructor to accept the extra data as
            kwargs.*

    Returns:
        decorator: A class decorator for registering a schema.

    Raises:
        AttributeError: If a constraint is provided that does not correspond to
            a method on the format.
    """
    def wrapper(schema):
        model_type = model or schema
        if issubclass(model_type, dict):
            format_type = DictFormat
        else:
            format_type = ObjectFormat

        class SchemaFormat(format_type):
            schema_class = schema
        SchemaFormat.model_type = model_type
        SchemaFormat.add_constraints(constraints, name=format_type.__name__)
        SchemaFormat.init()
        FormatRegistry.register_format(SchemaFormat)
        return schema
    return wrapper


def type(type=None, format=None, **constraints):
    """
    Build a type. The provided constraints are added to the format. The format
    is used when the new type is loaded or dumped.

    Arguments:
        type (type, optional): The type to add constraints to. If omitted, a
            format must be provided.
        format (type, optional): The format to add constrains to. If provided,
            it ignores the type argument. Defaults to the default format for
            the type. If omitted, a type must be provided.
        **constraints (Any): Constraints to add to the format. The kwarg names
            must be names of functions that are defined on the format class.
            The value of the kwarg is passed to the constraint function as its
            only argument.

    Returns:
        type: A new type suitable for use as a type annotation and in a schema.

    Raises:
        TypeError: If a type is provided that does not have a default format.
        AttributeError: If a constraint is provided that does not correspond to
            a method on the format.
    """
    format = format or FormatRegistry.get(type)

    class ConstrainedFormat(format):
        pass
    ConstrainedFormat.init()
    ConstrainedFormat.add_constraints(constraints, name=format.__name__)
    return FormatRegistry.build_type(ConstrainedFormat)


def load(data, type):
    """
    Deserialize data.

    Arguments:
        data (Any): The data to deserialize.
        type (type): The type to load the data into.

    Returns:
        Any: An instance of the type containing the deserialized data.

    Raises:
        TypeError: If the type does not have a format.
        TypeError: If any values are missing or have the wrong type.
        ValueError: If any values fail content constraints.
    """
    format = FormatRegistry.get(type)
    data, errors = format.load(data)
    if errors:
        raise TypeError(errors)
    errors = format.validate_content(data)
    if errors:
        raise ValueError(errors)
    return data


def dump(object, view=None):
    """
    Serialize data.

    Arguments:
        object (Any): An object to serialize.
        view (type, optional): The view to dump the object from. This allows
            defining an alternate format that serializes a data differently
            than the type that was used to construct the object. Defaults to
            the object's type.

    Returns:
        Any: The serialized data.

    Raises:
        TypeError: If the view does not have a format.
    """
    view = view or object.__class__
    format = FormatRegistry.get(view)
    return format.dump(object)


def validate(type):
    """
    Add custom validation. The provided function is used to validate the
    content when the type is loaded.

    Arguments:
        type (type): The type to validate.

    Returns:
        decorator: A function decorator for registering a validator.

    Raises:
        TypeError: If the type does not have a format.
    """
    format = FormatRegistry.get(type)

    def wrapper(validator):
        format.add_validator(validator)
        return validator
    return wrapper


__all__ = ['Object', 'type', 'schema', 'validate', 'dump', 'load']



import typing
from ..format_registry import FormatRegistry
from .raw_format import RawFormat


class List(RawFormat):
    formatted_type = list
    model_type = list
    generic_type = typing.List
    item_format = None

    @classmethod
    def from_type(self, list_type):
        item_type = list_type.__args__[0]

        class SpecificList(self):
            item_format = FormatRegistry.get(item_type)

        return SpecificList

    @classmethod
    def validate_type(self, data):
        error = super().validate_type(data)
        if error:
            return error

    @classmethod
    def validate_content(self, data):
        errors = [
            self.item_format.validate_content(item)
            for item in data
        ]
        return [error for error in errors if error]

    @classmethod
    def load(self, data):
        error = self.validate_type(data)
        if error:
            return None, error
        result = [
            self.item_format.load(item)
            for item in data
        ]
        errors = {
            index: error
            for index, (_, error) in enumerate(result)
            if error
        }
        if errors:
            return None, errors
        data = [value for value, _ in result]
        return data, None

    @classmethod
    def dump(self, data):
        return [
            self.item_format.dump(item)
            for item in data
        ]



from typing import get_type_hints
from ..format_registry import FormatRegistry
from .base_format import BaseFormat


class Object(BaseFormat):
    model_type = object
    formatted_type = dict
    schema_class = object
    type_hints = None
    child_formats = None

    @staticmethod
    def get_model_value(data, key):
        return getattr(data, key)

    @classmethod
    def init(self):
        super().init()
        type_hints = get_type_hints(self.schema_class)
        self.child_formats = [
            (name, FormatRegistry.get(type))
            for name, type in type_hints.items()
        ]
        self.formatted_names = {
            name: format.formatted_name
            if hasattr(format, 'formatted_name')
            else name
            for name, format in self.child_formats
        }
        self.defaults = {
            name: getattr(self.schema_class, name)
            for name in type_hints
            if hasattr(self.schema_class, name)
        }

    @classmethod
    def validate_extra_error(self, data):
        return {
            key: 'This key is not expected.'
            for key in data
            if key not in self.formatted_names.values()
        }

    @classmethod
    def validate_extra(self, data):
        return self.validate_extra_error(data)

    @classmethod
    def validate_extra_ignore(self, data):
        return {}

    @classmethod
    def get_extra(self, data):
        return self.get_extra_discard(data)

    @classmethod
    def get_extra_discard(self, data):
        return {}

    @classmethod
    def get_extra_keep(self, data):
        return {
            key: value
            for key, value in data.items()
        }

    @classmethod
    def validate_type(self, data: formatted_type):
        type_error = super().validate_type(data)
        if type_error:
            return type_error
        errors = self.validate_extra(data)
        for key, format in self.child_formats:
            data_key = self.formatted_names.get(key)
            if data_key not in data and key not in self.defaults:
                error = 'This key was not found, but it is required.'
            else:
                error = None
            if error:
                errors[key] = error
        return errors

    @classmethod
    def validate_content(self, data):
        errors = {}
        for key, format in self.child_formats:
            value = self.get_model_value(data, key)
            error = format.validate_content(value)
            if error:
                errors[key] = error
        return errors or super().validate_content(data)

    @classmethod
    def load(self, data: formatted_type) -> model_type:
        errors = self.validate_type(data)
        if errors:
            return None, errors
        init_kwargs = self.get_extra(data)
        for key, format in self.child_formats:
            data_key = self.formatted_names.get(key)
            if data_key in data:
                value, error = format.load(data[data_key])
                if error:
                    errors[key] = error
                else:
                    init_kwargs[key] = value
            else:
                init_kwargs[key] = self.defaults.get(key)
        if errors:
            return None, errors
        return self.model_type(**init_kwargs), None

    @classmethod
    def dump(self, data: model_type) -> formatted_type:
        formatted_data = self.get_extra(data)
        for key, format in self.child_formats:
            value = self.get_model_value(data, key)
            key = self.formatted_names.get(key)
            formatted_data[key] = format.dump(value)
        return formatted_data

    @classmethod
    def ignore_extra(self, enabled):
        if enabled:
            self.validate_extra = self.validate_extra_ignore
        else:
            self.validate_extra = self.validate_extra_error

    @classmethod
    def keep_extra(self, enabled):
        self.ignore_extra(enabled)
        if enabled:
            self.get_extra = self.get_extra_keep
        else:
            self.get_extra = self.get_extra_discard



from datetime import datetime
from .base_format import BaseFormat


class DateString(BaseFormat):
    model_type = datetime
    formatted_type = str
    formatted_encoding = None

    @classmethod
    def encoding(self, encoding: str):
        self.formatted_encoding = encoding

    @classmethod
    def load(self, data: formatted_type) -> model_type:
        error = self.validate_type(data)
        if error:
            return None, error
        try:
            return datetime.strptime(data, self.formatted_encoding), None
        except ValueError:
            error = 'Expected the encoding to be {}, but found {}.'.format(
                self.formatted_encoding,
                data,
            )
            return None, error

    @classmethod
    def dump(self, data: model_type) -> formatted_type:
        return data.strftime(self.formatted_encoding)



import typing
from ..format_registry import FormatRegistry
from .raw_format import RawFormat


class Union(RawFormat):
    generic_type = typing.Union

    @classmethod
    def from_type(self, union_type):
        class SpecificUnion(self):
            formatted_type = union_type
            model_type = union_type

        return SpecificUnion

    @classmethod
    def get_format(self, data):
        for type in self.formatted_type.__args__:
            format = FormatRegistry.get(type)
            error = format.validate_type(data)
            if not error:
                return format

    @classmethod
    def get_type_names(self):
        type_names = [
            type.__name__
            for type in self.formatted_type.__args__
        ]
        return ' or '.join(type_names)

    @classmethod
    def get_format_from_model_data(self, data):
        for type in self.formatted_type.__args__:
            format = FormatRegistry.get(type)
            if isinstance(data, format.model_type):
                return format
        message = 'Expected {}, but found {}.'.format(
            self.get_type_names(),
            data.__class__.__name__,
        )
        raise TypeError(message)

    @classmethod
    def load(self, data):
        format = self.get_format(data)
        if not format:
            error = 'Expected {}, but found {}.'.format(
                self.get_type_names(),
                data.__class__.__name__,
            )
            return None, error
        return format.load(data)

    @classmethod
    def dump(self, data):
        format = self.get_format_from_model_data(data)
        return format.dump(data)



from typing import List, Any
from .format_registry import FormatRegistry
from . import formats


integer_string = FormatRegistry.build_type(formats.IntegerString)


def setup():
    FormatRegistry.set_default(formats.NoneType)
    FormatRegistry.set_default(formats.Any)
    FormatRegistry.set_default(formats.String)
    FormatRegistry.set_default(formats.Float)
    FormatRegistry.set_default(formats.Integer)
    FormatRegistry.set_default(formats.DecimalString)
    FormatRegistry.set_default(formats.DateString)
    FormatRegistry.set_generic_default(formats.Union)
    FormatRegistry.set_generic_default(formats.List)
    FormatRegistry.set_default(formats.List.from_type(List[Any]), list)



from .object import Object


class Dict(Object):
    model_type = dict
    formatted_type = dict
    schema_class = object

    @staticmethod
    def get_model_value(data, key):
        return data[key]
